import atexit
from configparser import ConfigParser, SectionProxy
from os.path import exists
from os import listdir
from random import randint
from datetime import datetime, timedelta, timezone
import shutil

from .consts import CACHE_MAX, CACHE_PATH, CONFIG_PATH, TIME_FORMAT
from .utils import PackageType,absjoin
config=ConfigParser()
registries:SectionProxy | dict[str, str]
"""
レジストリの名前とそのインデックスのURLの辞書。

変更後は必ず、save_configを実行すること。
"""
_default_package_type:PackageType
def get_default_package_type()->PackageType:
	"""
	パッケージの既定の種類を取得する。

	Return
	---
	パッケージの既定の種類。
	"""
	return _default_package_type
def set_default_package_type(value:PackageType, auto_save:bool=True)->None:
	"""
	パッケージの既定の種類を設定する。
	
	Params
	---
	value: 設定する値。
	auto_save: 自動で保存するかどうか(既定値はTrue、Falseにした場合は、のちに、save_config()を実行すること)
	"""
	global _default_package_type
	_default_package_type=value
	if auto_save:save_config()
def save_config():
	"""
	コンフィグを保存する。
	コンフィグで保存されているものは、次の通り:
	* registries
	* default_package_type
	* cache_id
	"""
	config["registries"]=registries
	config["installation"]["default_package_type"]=_default_package_type.name
	config["installation"]["cache_id"]=str(cache_id)
	config["installation"]["cache_id_life"]=cache_id_life.strftime(TIME_FORMAT)
	with open(CONFIG_PATH,"w") as fp:
		config.write(fp)
if exists(CONFIG_PATH):
	config.read(CONFIG_PATH)
def get_cache_id():
	"""
	キャッシュIDを取得する。

	Return
	---
	現在のキャッシュID(これが変更されると、キャッシュが破棄される)。
	"""
	return cache_id
def clear_cache(auto_save:bool=True):
	"""
	キャッシュを破棄する。
	
	Params
	---
	auto_save: 自動で保存するかどうか(既定値はTrue、Falseにした場合は、のちに、save_config()を実行すること)
	"""
	global cache_id_life, cache_id
	cache_id_life=now+timedelta(1)
	cache_id=randint(0,CACHE_MAX)

	if auto_save:save_config()
@atexit.register
def delete_unused_cache():
	if exists(CACHE_PATH):
		[shutil.rmtree(absjoin(CACHE_PATH,d))for d in listdir(CACHE_PATH) if d!=str(cache_id)]
_tmp=False
registries=config["registries"]if "registries" in config else((_tmp:=True),{})[-1]
if not "installation"in config:config["installation"]=((_tmp:=True),{})[-1]
_default_package_type=PackageType[config["installation"]["default_package_type"]]if "default_package_type"in config["installation"]else((_tmp:=True),PackageType.py)[-1]
now=datetime.now(timezone.utc)
if "cache_id_life"not in config["installation"] or (cache_id_life:=datetime.strptime(config["installation"]["cache_id_life"],TIME_FORMAT))<now:
	if "cache_id"in config["installation"]:del config["installation"]["cache_id"]
	cache_id_life=now+timedelta(1)
cache_id=config["installation"].getint("cache_id")if"cache_id"in config["installation"]else((_tmp:=True),randint(0,CACHE_MAX))[-1]
if _tmp:save_config()