from dataclasses import dataclass, field
from enum import IntEnum
from os.path import abspath,join
import re
def absjoin(*pathes:str)->str:return abspath(join(*pathes))
def version_key(v:str)->tuple[int, int, int, bool, str]:
	tmp=v.split(".")
	tmp__=re.match(r"[0-9]*",tmp[2])
	if tmp__ is None:raise ValueError()
	tmp_=tmp__.group()
	mj,mn,pt=(int(i)for i in (*tmp[0:2],(tmp_)))
	suffix=tmp[2][len(tmp_):]+"".join(tmp[3:])
	return (mj,mn,pt,not suffix,suffix)
def escape_name(res:str)->str:
	res=res.replace('-' ,"-h")
	res=res.replace('\\',"-b")
	res=res.replace('/' ,"-s")
	res=res.replace(':' ,"-c")
	res=res.replace('*' ,"-a")
	res=res.replace('?' ,"-q")
	res=res.replace('"' ,"-d")
	res=res.replace('<' ,"-l")
	res=res.replace('>' ,"-g")
	res=res.replace('|' ,"-v")
	return res
class PackageType(IntEnum):
	"""
	パッケージの種類
	"""
	bin=0
	"""binパッケージ(旧来のバイナリ形式パッケージ)"""
	py=1
	"""pyパッケージ(pythonで記述されたパッケージ)"""
	auto=2
	"""パッケージの種類が指定されていないことを示すためのもの"""
@dataclass
class PackageInfo:
	"""
	パッケージ情報(パッケージのバージョンや種類など)
	"""
	version:str
	package_type:PackageType=PackageType.auto
	def __str__(self) -> str:
		return self.version if self.package_type==PackageType.auto else f"@{self.package_type.name}:{self.version}"
	@classmethod
	def read(cls,s:str):
		match s[0]:
			case "@":
				tmp=s.index(":")
				return PackageInfo(s[tmp+1:],PackageType[s[1:tmp]])
			case _ if s[0].isdigit():
				return PackageInfo(s)
			case _:
				raise ValueError()
@dataclass
class PackageMetaData:
	"""
	パッケージのメタ情報(パッケージの日本語名、紹介文など)
	"""
	name:dict[str,str]
	"""
	各言語でのパッケージの名前を格納する辞書で、必須。
	ロケール(ja-jpなど)をキー、名前を値として格納する。
	"""
	screenshot_count:int
	"""
	スクリーンショットの枚数で、必須。
	"""
	icon_sizes:list[int]
	"""
	アイコンの大きさの配列で、必須。
	"""
	authors:list[str]|None=None
	"""
	作者の配列。
	"""