import copy
import json
import os
import random
import shutil
import subprocess
import tkinter as tk
import tkinter.font as font
import tkinter.messagebox as messagebox
import tkinter.simpledialog as simpledialog
import tkinter.ttk as ttk
import unicodedata
import webbrowser

from ..maiwashi import absjoin
from .heart import *

ROOT_DIR=os.path.dirname(__file__)

root = tk.Tk()
root.title('ウミヘビ-面白いものを作ろう!')
root.state('zoomed')
root.iconbitmap(default=absjoin(ROOT_DIR,'Umihebi.ico'))
style = ttk.Style()
style.theme_use('classic')

def bindAllChildren(widget, sequence, func):
	widget.bind(sequence, func)
	for child in widget.winfo_children():
		bindAllChildren(child, sequence, func)

menu = tk.Frame(root, padx=12, bg='#111')
menu.pack(fill=tk.X)
menuFontJP = font.Font(
	family    ='游明朝',
	size      =12,
	weight    =font.NORMAL,
	slant     ="roman",
	underline =False,
	overstrike=False,
)
menuFontJP_ = font.Font(
	family    ='游明朝',
	size      =12,
	weight    =font.NORMAL,
	slant     ="roman",
	underline =True,
	overstrike=False,
)
def createMenuButton(text, bindClicked=None, parent=None):
	button = tk.Label(menu if parent is None else parent, text=text, fg='#ddd', bg=menu['bg'], font=menuFontJP, padx=10)
	if bindClicked is not None:
		button['cursor'] = 'hand2'
		button.bind('<Enter>', lambda e: button.configure(bg='#222', font=menuFontJP_))
		button.bind('<Leave>', lambda e: button.configure(bg=menu['bg'], font=menuFontJP))
		button.bind('<ButtonRelease-1>', lambda _: bindClicked())
	return button

webSiteLink = tk.Label(menu, text='https://ubazame.duckdns.org/Umihebi', cursor='hand2',
    font=('Sylfaen', 12, 'underline'), fg='#f55', bg=menu['bg'])
webSiteLink.bind('<ButtonRelease-1>', lambda event: webbrowser.open(event.widget['text'], autoraise=True))
webSiteLink.pack(side=tk.RIGHT)
tk.Label(menu, text='公式サイトはこちら→', font=('游明朝', 11),
	fg='#ddd', bg=menu['bg']).pack(side=tk.RIGHT)

loadingProjectName = tk.Entry(menu, border=1, relief=tk.FLAT,
	foreground='#ddd', width=12, justify=tk.RIGHT,
	readonlybackground='#222', state="readonly",
	font=('Sylfaen',11))
loadingProjectName.pack(side=tk.LEFT, padx=(0,4))

saveData = {
	"system": {
		"version": "0.2.0"
	},
	"programs": []
}
def saveThisProgram():
	global saveData, codes, savingStatus
	savingStatus['text'] = '保存しています...'
	staticEventSubscriptionNodes = []
	for code in codes:
		# 値渡しと参照渡しはきちんと区別しよう --- >_<
		code_dcopy = copy.deepcopy(code)
		if code['type'] == 'dec_observe':
			staticEventSubscriptionNodes.append(code_dcopy)
		elif code['type'] == 'exp_funcall':
			staticEventSubscriptionNodes[-1]['body'].append(code_dcopy)
	saveData['programs'] = staticEventSubscriptionNodes
	with open(absjoin(ROOT_DIR,f'heart/projects/{escapeProjectName(loadingProjectName.get())}/data.json'), 'w+', encoding='utf-8') as saveFile:
		json.dump(saveData, saveFile, indent=2)
	savingStatus['text'] = '保存しました!'
	root.after(700, lambda: savingStatus.configure(text=''))
createMenuButton('この作品を保存する', saveThisProgram).pack(side=tk.LEFT, fill=tk.X, ipady=4)

savingStatus = tk.Label(menu, text='', fg='#ddd', bg='#333', font=menuFontJP)
savingStatus.pack(side=tk.LEFT, fill=tk.Y)

# loadDialog と settingDialog の共通クラスを作ろうか迷っています。もう一枚くらいダイアログが入用になったら作るかも。 --- >_<, 2023/11/10

def closeLoadDialog():
	global loadDialog
	loadDialog.withdraw()
loadDialog = tk.Toplevel(root, bg='#111', width=600, height=600, padx=10, pady=10)
loadDialog.title('作品を読み込もう')
loadDialog.protocol('WM_DELETE_WINDOW', closeLoadDialog)
loadDialog.geometry(f"{loadDialog['width']}x{loadDialog['height']}+{(root.winfo_screenwidth()-loadDialog['width'])//2}+50")

currentlySelectingProject = None
def currentlySelectingProject_set(projectName:str):
	global currentlySelectingProject
	currentlySelectingProject = projectName

tk.Label(loadDialog, text='作品を読み込もう', fg='#ddd', bg=loadDialog['bg'], font=('游明朝',24)).pack()
tk.Frame(loadDialog, height=1, bg='#444').pack(pady=(10,20), fill=tk.X)
(loadDialog_projects:=tk.Frame(loadDialog, bg=loadDialog['bg'])).pack(fill=tk.X)
projectDirPath = absjoin(ROOT_DIR, 'heart/projects/')
(openAndClose:=tk.Frame(loadDialog, bg=loadDialog['bg'])).pack(side=tk.BOTTOM)
(opener:=tk.Label(openAndClose, text='読み込む', font=('游明朝',18), cursor='hand2', padx=12, pady=4)).pack(side=tk.RIGHT, padx=(4,0))
opener.bind('<ButtonRelease-1>', lambda e: [
	None if currentlySelectingProject is None else loadProject(currentlySelectingProject),
	closeLoadDialog()
	])
(closer:=tk.Label(openAndClose, text='やっぱやめとく', fg='#ddd', bg='#222', font=('游明朝',18), cursor='hand2', padx=8, pady=2)).pack(side=tk.RIGHT)
closer.bind('<ButtonRelease-1>', lambda e: closeLoadDialog())
loadDialog.withdraw()

# [hack] lambda文内で代入を行うためだけに作成 --- >_<, 2023/10/31
def substitution(TO, FROM):
	TO = FROM
def reloadLoadDialog():
	for oldProjectWidget in loadDialog_projects.winfo_children():
		oldProjectWidget.destroy()
	for project in os.listdir(projectDirPath):
		if os.path.isdir(os.path.join(projectDirPath, project)):
			(project:=tk.Label(loadDialog_projects, text=unescapeProjectName(project), fg='#ddd', bg=loadDialog['bg'], font=('游明朝',14), cursor='hand2')).pack(fill=tk.X)
			project.bind('<Enter>', lambda event: event.widget.configure(bg='#222' if event.widget['bg'] == loadDialog['bg'] else event.widget['bg']))
			project.bind('<Leave>', lambda event: event.widget.configure(bg=loadDialog['bg'] if event.widget['bg'] == '#222' else event.widget['bg']))
			project.bind('<ButtonRelease-1>', lambda event: [
				[project.configure(bg=loadDialog['bg'], font=('游明朝',14)) for project in loadDialog_projects.winfo_children()],
				event.widget.configure(bg='#333', font=('游明朝',14,'italic')),
				currentlySelectingProject_set(escapeProjectName(event.widget['text'])),
				opener.configure(fg='#ddd', bg='#22f')
			])
		else:
			raise FileExistsError(f'{projectDirPath}にファイルが含まれている可能性があります。')
def openLoadDialog():
	opener.configure(fg='#222', bg='#222')
	reloadLoadDialog()
	loadDialog.deiconify()
createMenuButton('作品を読み込む', openLoadDialog).pack(side=tk.LEFT, fill=tk.X, ipady=4)

def openPrompt2createNewProject():
	newProjectName = simpledialog.askstring(
		'作品を名付ける',
		f'''\t{
		random.choice(('"さあ、この世界を作ろう。" --- >_<','"他人の作った作品で遊んでいるだけじゃダメなんだ。自分で作品を作って遊ばないと。僕はそのためにウミヘビを創ったんだ。" --- >_<'))
		}\n\n新しく作る作品の名前を教えてくださいな。''')
	if newProjectName is None or newProjectName == '':
		return
	elif escapeProjectName(newProjectName) in os.listdir(absjoin(ROOT_DIR,'heart/projects')):
		messagebox.showinfo('あれ、聞いたことがあるぞ。',
			f'\t"おや、その名前は...。" --- >_<\n\n「{newProjectName}」は既にあなたの作品の中に存在します。\nこのままだとどっちがどっちか分からなくなってしまうので、別の名前を付けてあげてくださいな。')
		openPrompt2createNewProject()
	elif newProjectName == '>_<':
		# よく見つけたね。イースターエッグを仕込んでおいたよ。 --- >_<, 2023/11/14
		messagebox.showinfo('おや、僕を呼んだかい。>_<だよ。',
			f'僕は君の作品では......ないはず。だよね？\n君の新しい作品に幸あれかし。 :)\n\n\t>_<')
		openPrompt2createNewProject()
	else:
		shutil.copytree(absjoin(ROOT_DIR,'heart/new_project'), absjoin(ROOT_DIR,f'heart/projects/{escapeProjectName(newProjectName)}'))
		messagebox.showinfo(f'{newProjectName}を作ったよ！',
			f'「{newProjectName}」を作ったよ！\n{newProjectName}はあなたのお気に召すまま。')
		loadProject(escapeProjectName(newProjectName))
createMenuButton('新しい作品を作る', bindClicked=openPrompt2createNewProject).pack(side=tk.LEFT, fill=tk.Y)

def openPrompt2deleteThisProject():
	res = messagebox.askquestion(
		'[確認] 削除してもいいんですか？',
		f'''「{loadingProjectName.get()}」を削除してもいいですか？\n一度削除してしまうと、元に戻すことはできません。\n\n{
			random.choice(('"自分の作品を自分で壊すのは複雑な気分だね。" --- >_<', '"新しいものを作るためには、まず、古いものを壊さなくてはならない。"---ski.com'))
		}''')
	if res == 'yes':
		shutil.rmtree(absjoin(ROOT_DIR,f'heart/projects/{escapeProjectName(loadingProjectName.get())}'))
		messagebox.showinfo('削除しました。', f'「{loadingProjectName.get()}」を削除しました。')
		loadProject(os.listdir(absjoin(ROOT_DIR,'heart/projects'))[0])
(projectDeleterParent:=tk.Frame(menu,bg=menu['bg'])).pack(side=tk.LEFT, fill=tk.Y)
projectDeleter = createMenuButton('この作品を削除', bindClicked=openPrompt2deleteThisProject, parent=projectDeleterParent)

# 設定の設定
settingDialog = tk.Toplevel(root, bg='#111', width=600, height=600, padx=10, pady=10)
settingDialog.title('あなたのウミヘビの設定')
tk.Label(settingDialog, text='あなたのウミヘビの設定', fg='#ddd', bg=settingDialog['bg'], font=('游明朝',24)).pack()
tk.Frame(settingDialog, height=1, bg='#444').pack(pady=(10,20), fill=tk.X)
(returnTypeLabelSetting:=tk.Frame(settingDialog, bg=settingDialog['bg'])).pack(fill=tk.X, padx=30)
class tk_ToggleSwitch():
	def __init__(self, parent, state=False):
		self.state = state
		self.base  = tk.Frame(parent, bg='#333', padx=2, pady=2)
		self.on    = tk.Label(self.base, text='有効', bg='green', font=('游明朝',14), cursor='hand2')
		self.off   = tk.Label(self.base, text='無効', font=('游明朝',14), cursor='hand2')
		self.on.bind( '<ButtonRelease-1>', lambda e: self.switch())
		self.off.bind('<ButtonRelease-1>', lambda e: self.switch())
		self.on.pack(side=tk.LEFT)
		self.off.pack(side=tk.LEFT, fill=tk.Y)
		self.bindFunc = {}
		self.switch(state)
	def switch(self, state=None):
		if state is None:
			self.state = not self.state
		else:
			self.state = state
		if self.state:
			self.off['bg'] = self.base['bg']
			self.off['fg'] = self.base['bg']
			self.on['bg'] = 'green'
			self.on['fg'] = '#fff'
		else:
			self.on['bg'] = self.base['bg']
			self.on['fg'] = self.base['bg']
			self.off['bg'] = '#555'
			self.off['fg'] = '#fff'
		if '<AfterSwitch>' in self.bindFunc:
			self.bindFunc['<AfterSwitch>']()
	def bind(self, sequence, func):
		if sequence == '<AfterSwitch>':
			self.bindFunc[sequence] = func
		else:
			print(f'{sequence} は使えません')

setting = {}

(returnTypeLabelSettingToggle := tk_ToggleSwitch(returnTypeLabelSetting))
returnTypeLabelSettingToggle.base.pack(side=tk.RIGHT, anchor=tk.N)
def synchroSetting():
	global setting
	setting['simpleReturnTypeLabel'] = returnTypeLabelSettingToggle.state
returnTypeLabelSettingToggle.bind('<AfterSwitch>', synchroSetting)
tk.Label(returnTypeLabelSetting, text='ブロックの戻り値表示がウザいので短縮する', fg='#ddd', bg=returnTypeLabelSetting['bg'], font=('游明朝',14)).pack(anchor=tk.W)
tk.Label(returnTypeLabelSetting, text='多少反映に時間がかかります。\nすぐに反映させたいそこの君はウミヘビを再起動してね！', fg='#ddd', bg=returnTypeLabelSetting['bg'],
	font=('游明朝',10), justify=tk.LEFT).pack(anchor=tk.W)

settingDialog.protocol('WM_DELETE_WINDOW', lambda: settingDialog.withdraw())
settingDialog.geometry(f"{loadDialog['width']}x{loadDialog['height']}+{(root.winfo_screenwidth()-loadDialog['width'])//2}+50")
settingDialog.withdraw()
def openSettingDialog():
	settingDialog.deiconify()
createMenuButton('設定', openSettingDialog).pack(side=tk.LEFT, fill=tk.Y)



panelParent = tk.Frame(root, bg='green')
panelParent.pack(fill=tk.BOTH, expand=True)
panelParent.grid_rowconfigure(0, weight=1)
panelParent.columnconfigure(0, weight=1)
panelParent.columnconfigure(1, weight=0)

# blockArea & codeArea 's parent
codePanel = tk.Frame(panelParent)
codePanel.grid_propagate(False)
codePanel.grid(row=0, column=0, sticky=tk.NSEW)
codePanel.grid_rowconfigure(0, weight=1)
codePanel.columnconfigure(0, weight=1)
codePanel.columnconfigure(1, weight=1)

blocks = [
	{
		"type" : "dec_observe",
		"event": "BE9BADA0-030E-4AC8-8664-E5129DE50B6C",
		"body" : [],
		"args" : []
	},
	{
		"type" : "exp_funcall",
		"callee": {
			"type": "exp_ref_id",
			"id"  : "2919B826-2761-4D3A-970E-797DEEEBFECE"},
		"args": [{
			"type" : "exp_literal",
			"value": "Welcome to Umihebi!",
			"value_type": "string"
			}]
	},
	{
		"type" : "exp_funcall",
		"callee": {
			"type": "exp_ref_id",
			"id"  : "48CAE38C-F5D1-4629-9266-A9445E87E84D"},
		"args": [{
			"type" : "exp_literal",
			"value": "Welcome to Umihebi!",
			"value_type": "string"
			}]
	},
	{
		"type" : "exp_funcall",
		"callee": {
			"type": "exp_ref_id",
			"id"  : "14551C35-A20A-4D1D-88B8-E924A83A1FF0"},
		"args": [{
			"type" : "exp_literal",
			"value": "Welcome to Umihebi!",
			"value_type": "string"
			}]
	},
	{
		"type" : "stm_if",
		"condition": {
			"type" : "exp_literal",
			"value": None,
			"value_type": None},
		"body": []
	},
	{
		"type" : "stm_if",
		"condition": {
			"type": "exp_ref_id",
			"id"  : "14551C35-A20A-4D1D-88B8-E924A83A1FF0"},
		"body": [],
		"else": []
	},
	{
		# [beta] アラート系組み込み関数
		"type" : "exp_funcall",
		"callee": {
			"type": "exp_ref_id",
			"id"  : "E628135B-7E44-4B27-AA40-9EC88BE3BC08"},
		"args": [{
				"type" : "exp_literal",
				"value": "天気のあいさつ",
				"value_type": "string"
			}, {
				"type" : "exp_literal",
				"value": "こんにちは！明日は晴れるでしょうか？",
				"value_type": "string"
			}]
	}
]

'''	{
		"type" : "exp_funcall",
		"callee": {
			"type": "exp_ref_id",
			"id"  : "CA271D0A-19B1-4304-98BE-E1EDFD839879"},
		"args": [
			{
				"type" : "exp_literal",
				"value": "1",
				"value_type": "int"
			},
			{
				"type" : "exp_literal",
				"value": "1",
				"value_type": "int"
			}]
	},
	{
		"type": "dec_var",
		"name": "変数1",
		"id"  : "???",
		"var_type": {
			"type": "exp_ref_id",
			"id"  : "BF838FA8-705F-4246-8C22-C9B9ADDB6793"
			}
	},'''

blockArea = tk.Frame(codePanel, bg='#444', padx=10, pady=10)
blockArea.pack_propagate(False)
blockArea.grid(row=0, column=0, sticky=tk.NSEW)

draggingBlockWidget = None
dragStartPosition = {'x': None, 'y': None}
dragStartEventPosition = {'x': None, 'y': None}
draggingBlockInsertPosition = None

def dragStart(event, blockIndex):
	global draggingBlockWidget, dragStartPosition, dragStartEventPosition
	draggingBlockWidget = blockScript2widget(blocks[blockIndex], codePanel)
	selectedBlock = event.widget
	while str(blockArea).count('.!') != str(selectedBlock).count('.!') - 1:
		selectedBlock = selectedBlock._nametowidget(selectedBlock.winfo_parent())
	dragStartPosition['x'] = selectedBlock.winfo_rootx()
	dragStartPosition['y'] = selectedBlock.winfo_rooty() - codeArea.winfo_rooty()
	draggingBlockWidget.place(x=dragStartPosition['x'], y=dragStartPosition['y'])
	dragStartEventPosition['x'] = event.x; dragStartEventPosition['y'] = event.y
def dragging(event):
	global codesInfo, draggingBlockWidget, dragStartPosition, dragStartEventPosition, draggingBlockInsertPosition
	assert draggingBlockWidget is not None
	draggingBlockWidget.place(
		x = dragStartPosition['x'] + event.x - dragStartEventPosition['x'],
		y = dragStartPosition['y'] + event.y - dragStartEventPosition['y'])

	codeInsertPosition = None
	if event.x_root > codeArea.winfo_rootx():
		draggingBlockInsertPosition = None
		for loopi, codeInfo in enumerate(codesInfo):
			if codeInfo == []:
				continue
			if 400 > (codeInfo[0]['widget'].winfo_rootx()-draggingBlockWidget.winfo_rootx())**2 + (codeInfo[0]['widget'].winfo_rooty()-draggingBlockWidget.winfo_rooty())**2:
				codeInfo[0]['widget'].configure(width=12)
				draggingBlockInsertPosition = loopi
				hideCodeCursor()
				return
			else:
				codeInfo[0]['widget']['width'] = 4
		if event.y_root > codesOncodeArea()[-1].winfo_rooty() + codesOncodeArea()[-1].winfo_height():
			'''一番下にカーソルを表示する場合'''
			displayCodeCursor(len(codes))
		else:
			for loopi, code in enumerate(codesOncodeArea()):
				if (abs(code.winfo_rooty() - event.y_root + root.winfo_rooty()) < code.winfo_height() // 2):
					displayCodeCursor(loopi)
def dragFinish(event, blockIndex):
	global draggingBlockWidget, draggingBlockInsertPosition
	assert draggingBlockWidget is not None

	if draggingBlockInsertPosition is not None:
		codes[draggingBlockInsertPosition]['args'] = [blocks[blockIndex]]
		reloadCodeArea()
	if (
	draggingBlockWidget.winfo_rootx() + draggingBlockWidget.winfo_width() > codeCursor.winfo_rootx() and
	codeCursor.winfo_rootx() + codeCursor.winfo_width() > draggingBlockWidget.winfo_rootx() and
	draggingBlockWidget.winfo_rooty() + draggingBlockWidget.winfo_height() > codeCursor.winfo_rooty() and
	codeCursor.winfo_rooty() + codeCursor.winfo_height() > draggingBlockWidget.winfo_rooty()
	):
		insertCode(blocks[blockIndex], insertPosition)
	draggingBlockWidget.place_forget()
	draggingBlockWidget.destroy()
def bindDrag(widget, blockIndex):
	widget['cursor'] = 'hand2'
	widget.bind('<ButtonPress-1>'  , lambda event: dragStart(event, blockIndex))
	widget.bind('<B1-Motion>'      , dragging)
	widget.bind('<ButtonRelease-1>', lambda event: dragFinish(event, blockIndex))

codeMenu = tk.Frame(root, padx=2, pady=2, bg='#111')
index2delete = None
root.bind_all('<ButtonRelease-1>', lambda e: hideCodeMenu(), '+')
deleter = tk.Label(
	codeMenu, text='このブロックを削除する', font=('游明朝', 9),
	padx=10, pady=2, fg='#ccc', bg=codeMenu['bg'], cursor='hand2'
)
deleter.bind('<Enter>', lambda e: e.widget.config(background='#222'))
deleter.bind('<Leave>', lambda e: e.widget.config(background=codeMenu['bg']))
deleter.bind('<ButtonRelease-1>', lambda e: deleteCode(index2delete))
deleter.pack()
def showCodeMenu(event, codeIndex):
	global codeMenu, index2delete
	index2delete = codeIndex
	codeMenu.place(x=event.x_root, y=event.y_root - 21)
def hideCodeMenu():
	global codeMenu
	codeMenu.place_forget()
def bindCodeMenu(widget, codeIndex):
	widget.bind('<ButtonRelease-3>', lambda event: showCodeMenu(event, codeIndex))
	for child in widget.winfo_children():
		bindCodeMenu(child, codeIndex)

blockFontFamily = '游明朝'
blockFont = font.Font(
	family = blockFontFamily,
	size   = 12,
	weight = font.NORMAL,
	slant  = "roman",
	underline  = False,
	overstrike = False,
)
style.configure('block.TCombobox', foreground='#ccc', borderwidth=0, arrowsize=10)
style.map('block.TCombobox', background=[('readonly','#999')], fieldbackground=[('readonly','#333')],
	selectbackground=[('readonly', '#333')], selectforeground=[('readonly', '#ccc')])
root.option_add('*TCombobox*Listbox*Background', '#aaa')
root.option_add('*TCombobox*Listbox*selectForeground', '#ccc')
root.option_add('*TCombobox*Listbox*selectBackground', '#333')
def lengthDoubleByteStr(string:str)->int:
	"""
	全角・半角を区別して文字列の長さを返す

	https://ssrv.net/tech/python-strings-multi-byte/を参考にしました。ありがとうございます。 --- >_<, 2023/11/01
	"""
	count = 0
	for char in string:
		# 全角文字などの２バイト文字は'F','W','A'のいずれかに当てはまるとする
		count += 2 if unicodedata.east_asian_width(char) in 'FWA' else 1
	return count
def openVarNamePrompt(varNode):
	varNewName = simpledialog.askstring(
		f"変数「{blockNode['name']}」の名前を変える",
		f"変数「{blockNode['name']}」の新しい名前を教えてくださいな。\n\n※注意\n既に使用した名前は使えませんので気を付けて！\n")
	if varNewName is None or varNewName == '':
		return
	"""if newProjectName is None or newProjectName == '':
		return
	elif escapeProjectName(newProjectName) in os.listdir(absjoin(ROOT_DIR,'heart/projects')):
		messagebox.showinfo('あれ、聞いたことがあるぞ。',
			f'\t"おや、その名前は...。" --- >_<\n\n「{newProjectName}」は既にあなたの作品の中に存在します。\nこのままだとどっちがどっちか分からなくなってしまうので、別の名前を付けてあげてくださいな。')
		openPrompt2createNewProject()
	elif newProjectName == '>_<':
		# よく見つけたね。イースターエッグを仕込んでおいたよ。 --- >_<, 2023/11/14
		messagebox.showinfo('おや、僕を呼んだかい。>_<だよ。',
			f'僕は君の作品では......ないはず。だよね？\n君の新しい作品に幸あれかし。 :)\n\n\t>_<')
		openPrompt2createNewProject()
	else:
		shutil.copytree(absjoin(ROOT_DIR,'heart/new_project'), absjoin(ROOT_DIR,f'heart/projects/{escapeProjectName(newProjectName)}'))
		messagebox.showinfo(f'{newProjectName}を作ったよ！',
			f'「{newProjectName}」を作ったよ！\n{newProjectName}はあなたのお気に召すまま。')
		loadProject(escapeProjectName(newProjectName))"""
def blockScript2widget(blockNode, parent, blockidx=None):
	blockBorder = tk.Frame(parent, bg='#333')
	if (parent is not blockArea) and (parent is not codeArea) and (parent is not codePanel):
		blockBorder.configure(padx=1, pady=1)

	block = tk.Frame(blockBorder, bg='#111', padx=4, pady=4)
	block.pack(side=tk.LEFT)

	(curRow:=tk.Frame(block, bg=block['bg'])).pack() # 多層構造のブロックのための --- >_<, 2023/11/19

	blockInfo = None
	if parent is codeArea:
		blockInfo = []

	# ブロックに中身を詰めていきます。
	def addLabel2block(labelText):
		(label:=tk.Label(curRow, text=labelText, fg='#ddd', bg=block['bg'], font=blockFont)).pack(side=tk.LEFT)
		if parent == blockArea:
			bindDrag(label, blocks.index(blockNode))
	def addTypeSelector2block():
		(typeSelector:=ttk.Combobox(block,
			justify=tk.RIGHT, width=6, style='block.TCombobox', cursor='sb_v_double_arrow',
			font=(blockFontFamily,12), state='readonly', takefocus=False,
			exportselection=False, values=('すべて','数','文字列','整数','小数','真偽','なし')
			)).pack(side=tk.LEFT)
		typeSelector.selection_clear()
		typeSelector.bind('<MouseWheel>', lambda _: 'break')
		typeSelector.bind('<<ComboboxSelected>>', lambda e: typeSelector.selection_clear())
		typeSelector.current(0)
	def addNamePrompt2block():
		(namePrompt:=tk.Frame(block, bg='#333')).pack(side=tk.LEFT)
		(name:=tk.Label(namePrompt, text=blockNode['name'], fg='#ddd', bg='#1f1f1f', font=blockFont, cursor='hand2')).pack(side=tk.LEFT, fill=tk.Y)
		(helper:=tk.Label(namePrompt, text='クリックして\n名前を変える', fg='#ddd', bg=namePrompt['bg'], font=('tkDefaeultFont',8), cursor='hand2')).pack(side=tk.LEFT)
		helper.bind('<ButtonRelease-1>', lambda e: openVarNamePrompt(blockNode))
	def displayReturnType(uuid):
		returnType = {
			'4CB0AE0D-8131-447E-9213-31A3826967A0': '文字列',
			'F0EC4639-5918-4A0E-BC27-1CFA530659DF': '数',
			'4636E2A6-1C73-4DE9-B7E7-FD7F98EB4C12': 'なし'
		}
		if uuid == '4636E2A6-1C73-4DE9-B7E7-FD7F98EB4C12':
			return
		if type(uuid) is tuple:
			res = ''
			for i in uuid:
				res += returnType[i] + '/'
			tk.Label(blockBorder, text=res[:-1], fg='#fff', bg='#676e12', font=blockFont).pack(side=tk.LEFT, fill=tk.Y)
		else:
			tk.Label(blockBorder, text=(':' if setting['simpleReturnTypeLabel'] else '戻り値:')+returnType[uuid], fg='#fff', bg='#666', font=blockFont).pack(side=tk.LEFT, fill=tk.Y)
	def addEntry2block(node):
		# node: entryと変更を関連づけるノード
		if node['type'] == 'exp_literal':
			entry = tk.Entry(curRow, border=2, relief=tk.FLAT,
				insertwidth=1, insertbackground='#999',
				foreground='#ddd', width=4, justify=tk.RIGHT,
				background='#333',
				font=('Sylfaen', 12))
			def makeEntryEasier(en):
				en['width'] = max(4, lengthDoubleByteStr(en.get()))
			def synchroofEntryChanges():
				node['value'] = entry.get()
			entry.bind('<FocusIn>' , lambda event: makeEntryEasier(entry))
			entry.bind('<FocusOut>', lambda event: event.widget.configure(width=4))
			if 'value' in node and node['value'] is not None:
				entry.bind('<KeyRelease>', lambda e: synchroofEntryChanges())
				entry.insert(tk.END, node['value'])
			entry.pack(side=tk.LEFT)
			if blockInfo is not None:
				blockInfo.append({
					'argidx': 0,
					'blockidx': blockidx,
					'widget': entry
				})
		elif node['type'] == 'exp_funcall':
			blockScript2widget(node, block).pack(side=tk.LEFT)
	def newRow():
		(curRow:=tk.Frame(block, bg=block['bg'])).pack()

	if blockNode['type'] == 'dec_observe':
		addLabel2block("作品が実行されたとき")
	elif blockNode['type'] == 'dec_var':
		addTypeSelector2block()
		addLabel2block("型の変数")
		addNamePrompt2block()
		addLabel2block("を定義する")
	elif blockNode['type'] == 'stm_if':
		addLabel2block("もし")
		addEntry2block(blockNode['condition'])
		addLabel2block("なら")
		addLabel2block("する")
		if 'else' in blockNode:
			addLabel2block("でなければ")
			addLabel2block("する")
	elif blockNode['type'] == 'exp_funcall':
		if blockNode['callee']['id'] == '2919B826-2761-4D3A-970E-797DEEEBFECE':
			addLabel2block("コンソールに")
			addEntry2block(blockNode['args'][0])
			addLabel2block("と出力する")
			displayReturnType('4636E2A6-1C73-4DE9-B7E7-FD7F98EB4C12')
		elif blockNode['callee']['id'] == '48CAE38C-F5D1-4629-9266-A9445E87E84D':
			addLabel2block("コンソールに")
			addEntry2block(blockNode['args'][0])
			addLabel2block("と出力して改行する")
			displayReturnType('4636E2A6-1C73-4DE9-B7E7-FD7F98EB4C12')
		elif blockNode['callee']['id'] == '14551C35-A20A-4D1D-88B8-E924A83A1FF0':
			addLabel2block("コンソールに")
			addEntry2block(blockNode['args'][0])
			addLabel2block("と出力して待つ")
			displayReturnType('4CB0AE0D-8131-447E-9213-31A3826967A0')
		elif blockNode['callee']['id'] == 'CA271D0A-19B1-4304-98BE-E1EDFD839879':
			addEntry2block(blockNode['args'][0])
			addLabel2block(" + ")
			addEntry2block(blockNode['args'][1])
			displayReturnType(('4CB0AE0D-8131-447E-9213-31A3826967A0', 'F0EC4639-5918-4A0E-BC27-1CFA530659DF'))
		elif blockNode['callee']['id'] == 'E628135B-7E44-4B27-AA40-9EC88BE3BC08':
			addLabel2block('題名を')
			addEntry2block(blockNode['args'][0])
			addLabel2block('、内容を')
			addEntry2block(blockNode['args'][1])
			addLabel2block('としてダイアログボックスを出す')
	
	if parent == codeArea:
		bindCodeMenu(block, codes.index(blockNode))

	if parent is codeArea:
		return (blockBorder, blockInfo)

	return blockBorder

# [注意] 静的イベント購読ノードも1つの要素として扱われ、bodyには何も入っていません。
# よく分からなかったら>_<に聞いてください。 --- >_<
codes = []
codesInfo = [] # 適当に作りました。 --- >_<

codeArea = tk.Frame(codePanel, bg='#555', padx=10, pady=10)
codeArea.pack_propagate(False)
codeArea.grid(row=0, column=1, sticky=tk.NSEW)
codeCursor = tk.Frame(codeArea, bg='#777', width=100, height=30)
def codesOncodeArea():
	res = []
	for code in codeArea.winfo_children():
		if code == codeCursor:
			continue
		res.append(code)
	return res
insertPosition = None
def displayCodeCursor(insertPosi):
	global insertPosition
	insertPosition = insertPosi
	codeCursor.pack_forget()
	if insertPosition == len(codes):
		codeCursor.pack_forget()
		codeCursor.pack(anchor=tk.W)
		return
	for loopi, code in enumerate(codesOncodeArea()):
		code.pack_forget()
	for loopi, code in enumerate(codesOncodeArea()):
		if loopi == insertPosition:
			codeCursor.pack(anchor=tk.W)
		code.pack(anchor=tk.W, pady=(10 if codes[loopi]['type'] == 'dec_observe' else 0, 0))
def hideCodeCursor():
	codeCursor.pack_forget()
def addCode(blockNode):
	code = copy.deepcopy(blockNode)
	codes.append(code)
	reloadCodeArea()
def insertCode(blockNode, insertPosition):
	code = copy.deepcopy(blockNode)
	codes.insert(insertPosition, code)
	reloadCodeArea()
def deleteCode(codeIndex):
	del codes[codeIndex]
	reloadCodeArea()
def reloadCodeArea():
	global codesInfo
	# clean up
	for codeWidget in codesOncodeArea():
		codeWidget.destroy()
	codesInfo = []
	# redraw
	for blockNode in codes:
		codeBlock, info = blockScript2widget(blockNode, codeArea)
		codeBlock.pack(anchor=tk.W, pady=(10 if blockNode['type'] == 'dec_observe' else 0, 0))
		codesInfo.append(info)
	displayCodeCursor(len(codes))

runPanel = tk.Frame(panelParent, bg='#666')
runPanel.grid(row=0, column=1, sticky=tk.NSEW)

runner = tk.Frame(runPanel, bg='#333', width=180, height=180, padx=10, pady=10, cursor='hand2')
runner.pack_propagate(False)
tk.Label(runner, text='実行!', fg='#ddd', bg=runner['bg'], font=('游明朝', 24), cursor='hand2').pack()
runState = tk.Label(runner, text='この作品を\n実行してみましょう！', fg='#ddd', bg=runner['bg'], font=('游明朝', 12),
	cursor='hand2')
runState.pack(side=tk.BOTTOM, pady=(0,10))
runner.pack(anchor=tk.W, padx=10, pady=10)
def runThisProject():
	saveThisProgram()
	subprocess.call(
		f'''{absjoin(ROOT_DIR,'tail/run.bat')} "{absjoin(ROOT_DIR,'heart/projects/'+escapeProjectName(loadingProjectName.get()))}" ''',
		shell=True)
bindAllChildren(runner, '<ButtonRelease-1>', lambda _: runThisProject())

def loadProject(projectName:str):
	global codes
	if len(os.listdir(absjoin(ROOT_DIR,f'heart/projects'))) <= 1:
		projectDeleter.pack_forget()
	else:
		projectDeleter.pack(side=tk.LEFT, fill=tk.Y)
	with open(absjoin(ROOT_DIR,f'heart/projects/{projectName}/data.json'), 'r', encoding='utf-8') as loadFile:
		saveData = json.load(loadFile)
		codes = []
		for staticEventSubscriptionNode in saveData["programs"]:
			addCode(blocks[0])
			for scopeBlockNode in staticEventSubscriptionNode["body"]:
				addCode(scopeBlockNode)

		loadingProjectName['state'] = tk.NORMAL
		loadingProjectName.delete(0, tk.END)
		loadingProjectName.insert(tk.END, unescapeProjectName(projectName))
		loadingProjectName['state'] = "readonly"

with open(absjoin(ROOT_DIR,f'setting.json'), 'r', encoding='utf-8') as loadFile:
	setting = json.load(loadFile)

	returnTypeLabelSettingToggle.switch(setting['simpleReturnTypeLabel'])

	if os.path.isfile(absjoin(ROOT_DIR,f'head/projects/{setting["_workingProject"]}/data.json')):
		loadProject(setting["_workingProject"])
	else:
		loadProject(os.listdir(absjoin(ROOT_DIR,'heart/projects'))[0])

for loopi, blockNode in enumerate(blocks):
	blockScript2widget(blockNode, blockArea).pack(anchor=tk.W, pady=(0, 8))

def whenRootClosed():
	with open(absjoin(ROOT_DIR,f'setting.json'), 'w', encoding='utf-8') as systemJson:
		setting['_workingProject'] = loadingProjectName.get()
		json.dump(setting, systemJson, indent=2)
	root.destroy()
root.protocol('WM_DELETE_WINDOW', whenRootClosed)

root.bind('<Control-s>', lambda e: saveThisProgram())
root.bind('<Control-r>', lambda e: runThisProject())
root.bind('<F5>'       , lambda e: runThisProject())

root.mainloop()