from dataclasses import dataclass, field
import os
import threading
import tkinter as tk
import tkinter.ttk  as ttk
import traceback
from typing import Any, Callable

from PIL import Image,ImageTk

from .imports import get_readme, get_meta, get_icons, uninstall, install, list_packages, list_versions, absjoin, PackageInfo, clear_cache, get_screenshots

def iwashi_main(_):
	DEFAULT_SCREEN_SHOT=Image.open(absjoin(os.path.dirname(__file__),"default/screenshot.png"))
	
	root = tk.Tk()
	root['bg'] = '#555'
	root.title('ウルメイワシ-ウミヘビ公式市場にようこそ。')
	root.state('zoomed')
	root.iconbitmap(default=absjoin(os.path.dirname(__file__),'Iwashi.ico'))
	
	fontFamily = '游明朝'
	selectedTabColor = '#111'
	style = ttk.Style()
	style.configure('TSeparator', background='#4e454a')
	
	
	tabs = tk.Frame(root, padx=12, bg=root['bg'])
	tabs.pack(fill=tk.X, pady=(5,0))
	
	pages = tk.Frame(root)
	pages.grid_rowconfigure(0, weight=1)
	pages.grid_columnconfigure(0, weight=1)
	pages.pack(fill=tk.BOTH, expand=True)
	
	offcialProjectPage = tk.Frame(pages)
	offcialProjectPage.grid(row=0, column=0, sticky=tk.NSEW)
	
	UmihebiProjectPage = tk.Frame(pages, bg=selectedTabColor, pady=30)
	tk.Label(UmihebiProjectPage, text='哀しいかな、まだウミヘビ作品は1つも公開されていません。', fg='#ddd', bg=UmihebiProjectPage['bg'], font=(fontFamily,24)).pack()
	UmihebiProjectPage.grid(row=0, column=0, sticky=tk.NSEW)
	
	offcialProjectPage.tkraise()
	
	projectsCanvas4scroll = tk.Canvas(offcialProjectPage,
		width=500, height=750,
		bd=0, highlightthickness=0, bg=selectedTabColor)
	projectsScrollBar = ttk.Scrollbar(offcialProjectPage, orient=tk.VERTICAL,
		command=projectsCanvas4scroll.yview)
	projectsCanvas4scroll.pack(side=tk.LEFT, anchor=tk.N, fill=tk.Y)
	projectsScrollBar.pack(side=tk.LEFT, fill=tk.Y)
	projectsCanvas4scroll.configure(yscrollcommand=projectsScrollBar.set)
	
	def _mouseWheelOnProjects(event):
		projectsCanvas4scroll.yview_scroll(-(event.delta // 120), 'units')
	
	projectsCanvas4scroll.update_idletasks()
	projects = tk.Frame(offcialProjectPage, bg=selectedTabColor)
	projects.bind('<Enter>',
		lambda _: root.bind_all('<MouseWheel>', _mouseWheelOnProjects)
	)
	projects.bind('<Leave>', lambda _: root.unbind_all('<MouseWheel>'))
	projects.bind(
		'<Configure>',
		lambda e: projectsCanvas4scroll.configure(
			scrollregion=(0,0,projectsCanvas4scroll['width'],projects.winfo_height())
		)
	)
	
	tk.Label(projects, text='もう既にあなたのコンピュータ内に棲み着いているもの', fg='#ddd', bg=projects['bg'], font=(fontFamily, 12)).pack(pady=(40,20))
	livingProjects = tk.Frame(projects, bg=projects['bg'])
	livingProjects.pack(fill=tk.X)
	# [hack] tk.Frame as ttk.Separator --- >_<
	tk.Frame(projects, bg="#333", height=1, bd=0).pack(fill=tk.X, padx=10, pady=20)
	
	tk.Label(projects, text='あなたのコンピュータの中に入りこむ機会を伺うもの', fg='#ddd', bg=projects['bg'], font=(fontFamily, 12)).pack(pady=20)
	waitingProjects = tk.Frame(projects, bg=projects['bg'])
	waitingProjects.pack(fill=tk.X, pady=(0,20))
	
	loadingStatus = tk.Label(projects, fg='#ddd', bg='#333', font=(fontFamily,12))
	loadingStatus.pack(fill=tk.X)
	
	clearCacheButton = tk.Label(projects, text='最新の状態に更新する', fg='#ddd', bg='#262626', font=(fontFamily,16), padx=8, pady=2, cursor='hand2')
	clearCacheButton.bind('<ButtonRelease-1>', lambda _:reloadProjects())
	clearCacheButton.pack(pady=(20,20))
	
	projects.pack(side=tk.LEFT, fill=tk.Y)
	projectsCanvas4scroll.create_window((0, 0), window=projects, width=projectsCanvas4scroll['width'], anchor=tk.NW)
	
	moreAbout = tk.Frame(offcialProjectPage, bg='#333', padx=30, pady=20)
	projectDetail = tk.Frame(moreAbout, bg=moreAbout['bg'])
	
	uninstallWarningBorder = tk.Frame(root, bg='#777')
	uninstall_warning = tk.Frame(uninstallWarningBorder, width=700, height=400, pady=20, bg='#222')
	tk.Label(uninstall_warning, text='本当にいいのかな？', fg='#ddd', bg=uninstall_warning['bg'], font=(fontFamily,24,'bold')).pack(pady=(0,30))
	tk.Label(uninstall_warning, text='この作品をPCから追い出していいんですか？\n', fg='#ddd', bg=uninstall_warning['bg'], font=(fontFamily,15)).pack()
	tk.Label(uninstall_warning, text='予測される不利益', fg='#ddd', bg=uninstall_warning['bg'], font=(fontFamily,15,"bold")).pack()
	predictedDisadvantages = tk.Label(uninstall_warning, bg=uninstall_warning['bg'], font=(fontFamily,15))
	predictedDisadvantages.pack()
	confirm = tk.Frame(uninstall_warning, bg=uninstall_warning['bg'])
	cancel = tk.Label(confirm, text='いや、やめておきます', fg='#ddd', bg='#555', font=(fontFamily,15), padx=8, pady=2, cursor='hand2')
	cancel.bind('<ButtonRelease-1>', lambda e: uninstallWarningBorder.place_forget())
	cancel.pack(side=tk.LEFT, padx=(0,20))
	ok     = tk.Label(confirm, text='うん、さよならだ', fg='#d22', bg='#111', font=(fontFamily,15), padx=9, pady=3, cursor='hand2')
	ok.bind('<ButtonRelease-1>', lambda e: uninstallWarningBorder.place_forget())
	ok.pack(side=tk.LEFT)
	confirm.pack(side=tk.BOTTOM)
	uninstall_warning.pack_propagate(False)
	uninstall_warning.pack(padx=2, pady=2)
	def startUninstall(package_name:str):
		errorHappened = False
		try:
			uninstall(package_name)
		except:
			traceback.print_exc()
			downloader['text'] = '失敗しました。キャッシュを破棄してもう一度試してみて下さい'
			errorHappened = True
		if errorHappened is not True:
			downloader['bg']   = '#3c3c3c'
			downloader['text'] = '✓完了'
			downloader['cursor'] = 'arrow'
			updateProjects()
		uninstallWarningBorder.place_forget()
	def showUninstallDialog(package_name:str):
		ok.pack(side=tk.LEFT)
		if package_name == 'umihebi':
			predictedDisadvantages['fg']   = '#d22'
			predictedDisadvantages['text'] = 'あの便利なウミヘビがもう一度ここで入手するまで使えなくなります。'
		elif package_name == 'maiwashi' or package_name == 'urume_iwashi':
			predictedDisadvantages['fg']   = '#d22'
			predictedDisadvantages['text'] = 'このソフトが壊れて使えなくなります。'
			ok.pack_forget()
		elif package_name == 'katakuchi_iwashi':
			predictedDisadvantages['fg']   = '#d22'
			predictedDisadvantages['text'] = 'イワシがCUIで使えなくなります。'
		else:
			predictedDisadvantages['fg']   = '#ddd'
			predictedDisadvantages['text'] = 'たぶんありません。'
		ok.bind('<ButtonRelease-1>', lambda e: threading.Thread(target=startUninstall, args=(package_name,), daemon=True).start())
		uninstallWarningBorder.place(x=(root.winfo_width() - uninstall_warning['width'])//2, y=100)
	
	versionTool = tk.Frame(projectDetail)
	
	downloader  = tk.Label(versionTool,
		fg='#ddd', bg=moreAbout['bg'], font=(fontFamily, 15), padx=6, pady=2, cursor='hand2')
	uninstaller = tk.Label(versionTool, text='コンピュータから追い出す',
		fg='#ddd', bg=moreAbout['bg'], font=(fontFamily, 15), padx=6, pady=2, cursor='hand2')
	uninstaller.bind('<Enter>', lambda e: uninstaller.configure(font=(fontFamily,15,'underline')))
	uninstaller.bind('<Leave>', lambda e: uninstaller.configure(font=(fontFamily,15)))
	def startDownload(package_name:str):
		errorHappened = False
		try:
			install(package_name, package_info=PackageInfo(list_versions(package_name)[-1]), is_update=False)
		except:
			traceback.print_exc()
			downloader['text'] = '！入手に失敗しました。キャッシュを破棄して試してみて！'
			errorHappened = True
		if errorHappened is not True:
			downloader['bg']   = '#3c3c3c'
			downloader['text'] = '✓入手済み'
			downloader['cursor'] = 'arrow'
			updateProjects()
	def startUpdate(package_name:str):
		errorHappened = False
		try:
			install(package_name, package_info=PackageInfo(list_versions(package_name)[-1]), is_update=True)
		except:
			traceback.print_exc()
			downloader['text'] = '！入手に失敗しました。キャッシュを破棄して試してみて！'
			errorHappened = True
		if errorHappened is not True:
			downloader['bg']   = '#3c3c3c'
			downloader['text'] = '✓入手済み'
			downloader['cursor'] = 'arrow'
			updateProjects()
	def setupVersionTool(package_name:str):
		downloader['bg']     = '#555'
		downloader['cursor'] = 'hand2'
		if projectsData[package_name].parent is livingProjects:
			if projectsData[package_name].version != projectsData[package_name].versionLatest:
				downloader.configure(fg='white', bg='sea green')
			uninstaller.bind('<ButtonRelease-1>',
				lambda e: showUninstallDialog(package_name))
			uninstaller.pack(side=tk.RIGHT)
			downloader['text'] = '最新の状態に更新'
			downloader.bind('<ButtonRelease-1>',
				lambda e: downloader.configure(text='インターネットから取得中...') == threading.Thread(target=startUpdate, args=(package_name,), daemon=True).start()
				)
		else:
			uninstaller.pack_forget()
			downloader['text'] = '入手してみる'
			downloader.bind('<ButtonRelease-1>',
				lambda e: downloader.configure(text='インターネットから取得中...') == threading.Thread(target=startDownload, args=(package_name,), daemon=True).start()
				)
	downloader.pack(side=tk.RIGHT, anchor=tk.N)
	
	versionTool.place(relx=1, rely=0, anchor=tk.NE)
	
	moreAbout_title   = tk.Label(projectDetail, fg='#ddd', bg=moreAbout['bg'], font=(fontFamily, 24, "bold"))
	moreAbout_version = tk.Label(projectDetail, fg='#ddd', bg=moreAbout['bg'], font=(fontFamily, 12))
	moreAbout_status  = tk.Label(projectDetail, fg='aquamarine', bg=moreAbout['bg'], font=(fontFamily, 12), justify=tk.LEFT)
	moreAbout_desc    = tk.Label(projectDetail, fg='#ddd', bg=moreAbout['bg'], font=(fontFamily, 12), justify=tk.LEFT)
	moreAbout_credit  = tk.Label(projectDetail, fg='#ddd', bg=moreAbout['bg'], font=(fontFamily, 12))
	moreAbout_title.pack(anchor=tk.W)
	moreAbout_version.pack(anchor=tk.W)
	moreAbout_status.pack(anchor=tk.W)
	moreAbout_desc.pack(anchor=tk.W)
	moreAbout_credit.pack(anchor=tk.W)
	
	projectDetail.pack(fill=tk.X)
	sepa = ttk.Separator(moreAbout)
	sepa.pack(fill=tk.X, pady=20)
	(moreAbout_screenshot:=tk.Label(moreAbout,
		font=('Courier',24,'bold'), fg='#ccc',
		bg=moreAbout['bg'], borderwidth=0, compound='bottom')).pack(fill=tk.BOTH, expand=True)
	moreAbout.pack(fill=tk.BOTH, expand=True)
	
	@dataclass
	class ProjectData:
		parent:tk.Frame
		AKA:dict[str,str]
		description:str="読み込み中..."
		credit:str=""
		version:str|None="???"
		versionLatest:str|None='???'
		thumbnail80x80:tk.PhotoImage|ImageTk.PhotoImage=tk.PhotoImage(file=absjoin(os.path.dirname(__file__),'default/icon80x80.png'))
		screenshotDirPath:str=absjoin(os.path.dirname(__file__),'default/Iwashi')
		screenshotCount:int=0
		screenshot:Image.Image=field(default_factory=lambda:DEFAULT_SCREEN_SHOT)
		widget:tk.Widget|None=None
		on_update:Callable[[],Any]=lambda:0
	projectsData:dict[str,ProjectData] = {}
	
	def reloadProjects():
		for projectBorder in livingProjects.winfo_children():
			projectBorder.destroy()
		for projectBorder in waitingProjects.winfo_children():
			projectBorder.destroy()
		root.after(1,lambda:(clear_cache(),loadOnlineOffcialProjects()))
	def updateProjects():
		for key, values in projectsData.items():
			projectdata=projectsData[key]
			l=list_packages()
			tmp=key in l
			projectdata.parent = livingProjects if tmp else waitingProjects
			projectdata.version=l[key].version if tmp else None
			values.on_update()
		if project_name is not None:focusSwitch(project_name)
	def configWidgetBg(widget:tk.Widget, bg):
		if widget.widgetName == 'frame' or widget.widgetName == 'label':
			widget['bg'] = bg
		for child in widget.winfo_children():
			configWidgetBg(child, bg)
	moreAbout_screenshot_current_image:ImageTk.PhotoImage
	project_name:str|None=None
	def focusSwitch(projectName:str):
		nonlocal moreAbout_screenshot_current_image,project_name
		for projectBorder in livingProjects.winfo_children()+waitingProjects.winfo_children():
			if projectBorder.winfo_children()[0] == projectsData[projectName].widget:
				projectBorder['bg'] = '#444'
			else:
				projectBorder['bg'] = projects['bg']
		setupVersionTool(projectName)
		moreAbout_title.config(text=projectsData[projectName].AKA['ja-jp'])
		#tmp=projectsData[projectName].version
		moreAbout_version.config(
			text='- 未入手 -' if projectsData[projectName].version is None else
				f'バージョン {projectsData[projectName].version},  最新版は {projectsData[projectName].versionLatest}')
		if projectsData[projectName].version != projectsData[projectName].versionLatest:
			moreAbout_status.config(text=
				 "更新データがあります。右上のボタンを押して、最新の状態に更新してあげてください。\n"
				f"(あなたの{projectsData[projectName].AKA['ja-jp']}を常にキレイに保つことができます。)")
		else:
			moreAbout_status.config(text='')
		moreAbout_desc.config(text=projectsData[projectName].description)       
		moreAbout_credit.config(text=projectsData[projectName].credit)
		img=projectsData[projectName].screenshot
		if img is DEFAULT_SCREEN_SHOT:
			moreAbout_screenshot.configure(text='NO IMAGE',
				image=(moreAbout_screenshot_current_image:=ImageTk.PhotoImage(DEFAULT_SCREEN_SHOT))
			)
		else:
			screenshotWidth = moreAbout_screenshot.winfo_width()
			screenshotHeight = moreAbout_screenshot.winfo_height()
			moreAbout_screenshot.config(
				text='',
				image=(moreAbout_screenshot_current_image:=ImageTk.PhotoImage(img.resize(min((screenshotWidth, int(img.height * screenshotWidth / img.width)),(int(img.width * screenshotHeight / img.height), screenshotHeight)))))
			)
		project_name=projectName
	
	def bindFocusSwitch(widget, projectName):
		widget['cursor'] = 'hand2'
		widget.bind('<ButtonRelease-1>', lambda e: focusSwitch(projectName))
		for child in widget.winfo_children():
			bindFocusSwitch(child, projectName)
	def addProject(name):
		cur_data=projectsData[name]
		projectBorder = tk.Frame(oldparent:=cur_data.parent, bg=projects['bg'])
		project = tk.Frame(projectBorder, height=100, bg=projects['bg'], pady=10)
		project.pack_propagate(False)
	
		(thumbnail:=tk.Label(project, width=80, height=80, borderwidth=0)).pack(side=tk.LEFT, padx=10)
		(titleSpace:=tk.Frame(project, bg=project['bg'])).pack(anchor=tk.W)
		(aka_jp    :=tk.Label(titleSpace, fg='#ddd', bg=project['bg'], font=(fontFamily, 15, "bold"))).pack(side=tk.LEFT)
		(needUpdate:=tk.Label(titleSpace, fg='aquamarine', bg=project['bg'], font=(fontFamily, 12))).pack(side=tk.LEFT)
	
		(desc  :=tk.Label(project, fg='#ddd', bg=project['bg'], font=(fontFamily, 12))).pack(anchor=tk.W)
		(credit:=tk.Label(project, fg='#ddd', bg=project['bg'], font=(fontFamily, 12))).pack(anchor=tk.W)
		project.bind('<Enter>', lambda e:[
			configWidgetBg(project, '#222'),
			projectBorder.configure(bg='#222') if projectBorder['bg'] == projects['bg'] else None
			])
		project.bind('<Leave>', lambda e: configWidgetBg(project, projects['bg']))
		bindFocusSwitch(project, name)
		project.pack(fill=tk.X, padx=(0,7))
		cur_data.widget = project
		projectBorder.pack(fill=tk.X)
		def updateProject():
			if oldparent!=projectsData[name].parent:
				projectBorder.destroy()
				addProject(name)
				return
			thumbnail.config(image=cur_data.thumbnail80x80)
			aka_jp.config(text=cur_data.AKA['ja-jp'])
			needUpdate['text'] = '- 更新があります'if cur_data.version != cur_data.versionLatest else''
	
			desc_short = cur_data.description.split('\n')[0]
			if desc_short[-1] == '\n':
				desc_short = desc_short[:-1]
			if len(desc_short) > 24:
				desc_short = desc_short[:22] + '...'
			desc.config(text=desc_short)
			credit.config(text=cur_data.credit)
		cur_data.on_update=updateProject
		updateProject()
	
	for key in projectsData:
		addProject(key)
	
	
	tabInfo = {
		'ウバザメ公式作品': {
			'tab' : None,
			'page': offcialProjectPage
		},
		'ウミヘビ作品市場 (準備中)': {
			'tab' : None,
			'page': UmihebiProjectPage
		}
	}
	def changeTabFocus(tabName):
		tabInfo[tabName]['page'].tkraise()
		for key in tabInfo:
			if key == tabName:
				tabInfo[key]['tab']['bg'] = selectedTabColor
			else:
				tabInfo[key]['tab']['bg'] = '#333'
	def initTab(tabName):
		tabInfo[tabName]['tab'] = tk.Label(tabs, text=tabName, fg='#ddd', bg='#333', font=(fontFamily, 12), padx=12, pady=2,
			cursor='hand2')
		tabInfo[tabName]['tab'].bind('<ButtonRelease-1>', lambda e: changeTabFocus(tabName))
		tabInfo[tabName]['tab'].pack(side=tk.LEFT, padx=1)
	for key in tabInfo:
		initTab(key)
	changeTabFocus('ウバザメ公式作品')
	
	# このへん関数名とんでもないな... --- >_<, 2023/11/19
	errorWhileLoading:bool
	errorWhileLoading_lock=threading.Lock()
	stopLoadingProjectAnime:bool
	total_packages_count:int
	loaded_packages_count:int
	loaded_packages_count_lock=threading.Lock()
	def loadOnlineOffcialProject(name:str,value:PackageInfo|None):
		projectsData[name] = cur_data = ProjectData(
			livingProjects if name in list_packages(include_registry=False) else waitingProjects,
			{'ja-jp': name},
			version=None if value is None else value.version,
			versionLatest=None if value is None else list_versions(name)[-1])
		addProject(name)
		thread=threading.Thread(target=lambda:_loadOnlineOffcialProject_main(name,cur_data))
		thread.start()
		return thread
	def _loadOnlineOffcialProject_main(name:str,cur_data:ProjectData):
		nonlocal loaded_packages_count,errorWhileLoading,loaded_packages_count
		try:
			cur_data.description = (lambda s:'説明はありません。' if s is None else s)(get_readme(name))
	
			metaData = get_meta(name)
			cur_data.AKA = metaData.name
			if metaData.authors is None:
				cur_data.credit = '不明 (>_<かski.comの可能性が高いです)'
			else:
				cur_data.credit = '、'.join(metaData.authors)
			cur_data.screenshotDirPath,tmp=get_screenshots(name)
			cur_data.screenshotCount=tmp
			if tmp:
				projectsData[name].screenshot = Image.open(f"{projectsData[name].screenshotDirPath}/1.png")
			icon_dir,icon_sizes=get_icons(name)
			if 80 in icon_sizes:
				cur_data.thumbnail80x80 = tk.PhotoImage(file=absjoin(icon_dir, f"x80.png"))
			elif icon_sizes:
				cur_data.thumbnail80x80 = ImageTk.PhotoImage(Image.open(absjoin(icon_dir, f"x{max(icon_sizes,key=lambda i:(not i%80,i))}.png")))# 80の倍数を抽出(おそらくキレイになる)
		except:
			traceback.print_exc()
			with errorWhileLoading_lock:
				if not errorWhileLoading:
					loadingStatus['text'] = f'{name}の読み込み中に問題が発生しました。\nもう一回このソフトを再起動すると自己修復できるかもしれません。'
					errorWhileLoading = True
		else:
			with loaded_packages_count_lock:loaded_packages_count+=1
			cur_data.on_update()
	
	def _loadOnlineOffcialProjects():
		nonlocal errorWhileLoading, stopLoadingProjectAnime,total_packages_count,loaded_packages_count
		errorWhileLoading = False
		tmp=list_packages(True)
		total_packages_count=len(tmp)
		loaded_packages_count=0
		threading.Thread
		for name,value in tmp.items():
			loadOnlineOffcialProject(name, value)
		_loadOnlineOffcialProject_loop()
	def _loadOnlineOffcialProject_loop():
		nonlocal stopLoadingProjectAnime
		if loaded_packages_count<total_packages_count and not errorWhileLoading:
			root.after(1,_loadOnlineOffcialProject_loop)
			return
		stopLoadingProjectAnime = True
		if errorWhileLoading is False:
			loadingStatus['text'] = 'すべての公式作品の読み込みが完了しました。'
	def loadingProjectsAnime():
		nonlocal stopLoadingProjectAnime
		if stopLoadingProjectAnime:return
		if loadingStatus['text'][-3:] == '.' * 3:
			loadingStatus['text'] = loadingStatus['text'][:-3]
		else:
			loadingStatus['text'] = loadingStatus['text'] + '.'
		root.after(500, loadingProjectsAnime)
	def loadOnlineOffcialProjects():
		nonlocal stopLoadingProjectAnime
		stopLoadingProjectAnime=False
		loadingStatus['text'] = 'インターネットから情報を取得しています'
		loadingProjectsAnime()
		threading.Thread(target=_loadOnlineOffcialProjects, daemon=True).start()
	
	root.after(200, loadOnlineOffcialProjects)
	
	root.mainloop()